library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;

entity main is
	port(
		clockIn: in std_logic;
		data: in std_logic_vector(1 downto 0);
		
		controlData: in std_logic;
		controlClock: in std_logic;
		
		outEnable: in std_logic;
		
		rawClockOut: out std_logic;
		out1: out std_logic;
		out2: out std_logic;
		clockOut: out std_logic
	);
end main;

architecture rtl of main is
	-- regs memory
	signal regs: unsigned(28 downto 0) := (others => '0');
	
	-- regs aliases
	signal div1LowCountReg: unsigned(4 downto 0);
	signal div1HighEndReg: unsigned(4 downto 0);
	signal div2HighStartReg: unsigned(4 downto 0);
	signal div2HighCountReg: unsigned(4 downto 0);
	signal prescaleCounterReg: unsigned(5 downto 0);
	signal delayDiv2Reg: std_logic;
	signal useDataInputReg: std_logic;
	signal prescaleBypassReg: std_logic;
	
	signal counter1: natural range 0 to 31 := 0;
	signal counter2: natural range 0 to 31 := 0;
	signal div1: std_logic := '0';
	signal div2: std_logic := '0';
	signal div2Delayed: std_logic := '0';
	signal latchedData: std_logic_vector(1 downto 0);
	signal outEnableLatch: std_logic := '0';
	signal prescaleEnableLatch: std_logic := '0';
	signal clock: std_logic := '0';
	signal prescaledClock: std_logic := '0';
	signal prescaleCounter: natural range 0 to 63 := 0;
	signal controlDataLatch: std_logic := '0';
	signal controlClockLatch: std_logic_vector(1 downto 0) := (others => '0');

begin

	-- set the control register bits
	controlProcess: process(clockIn)
	begin
		if rising_edge(clockIn) then
			controlDataLatch <= controlData;
			controlClockLatch <= controlClockLatch(0) & controlClock;
			if controlClockLatch = "01" then
				regs <= controlDataLatch & regs(28 downto 1);
			end if;
		end if;
	end process;
	div1LowCountReg <= regs(4 downto 0);
	div1HighEndReg <= regs(9 downto 5);
	div2HighStartReg <= regs(14 downto 10);
	div2HighCountReg <= regs(19 downto 15);
	delayDiv2Reg <= regs(20);
	useDataInputReg <= regs(21);
	prescaleCounterReg <= regs(27 downto 22);
	prescaleBypassReg <= regs(28);

	-- clock prescaler
	prescalerProcess: process(clockIn)
	begin
		if rising_edge(clockIn) then
			-- latch output enable signal with clockIn
			prescaleEnableLatch <= outEnable;
			
			-- generate prescaled clock
			if prescaleCounter = prescaleCounterReg then
				prescaleCounter <= 0;
				prescaledClock <= not prescaledClock;
			else
				prescaleCounter <= prescaleCounter + 1;
			end if;
			if prescaleEnableLatch = '0' then
				prescaleCounter <= 0;
				prescaledClock <= '0';
			end if;
		end if;
	end process;
	
	-- select prescaled clock or full speed
	clock <= clockIn when prescaleBypassReg = '1' else prescaledClock;
	
	-- generate div1 and div2
	divideProcess: process(clock)
	begin
		if rising_edge(clock) then
			-- latch buffered output enable signal with derived clock
			outEnableLatch <= prescaleEnableLatch;

			-- generate div1 signal
			div1 <= '0';
			if counter1 < div1LowCountReg then
				counter1 <= counter1 + 1;
			else
				div1 <= '1';
				if counter1 < div1HighEndReg then
					counter1 <= counter1 + 1;
				else
					counter1 <= 0;
					latchedData <= data;
				end if;
			end if;
			
			-- div2 signal starts at a fixed time offset from the start of div1 signal
			if counter1 = div2HighStartReg then
				div2 <= '1';
				counter2 <= 0;
			else
				if counter2 < div2HighCountReg then
					counter2 <= counter2 + 1;
				else
					div2 <= '0';
				end if;
			end if;

			if outEnableLatch = '0' then
				counter1 <= 0;
				counter2 <= 0;
				div1 <= '0';
				div2 <= '0';
			end if;
		end if;
	end process;
	
	-- half clock cycle delay for div2
	delay: process(clock)
	begin
		if falling_edge(clock) then
			div2Delayed <= div2;

			if outEnableLatch = '0' then
				div2Delayed <= '0';
			end if;
		end if;
	end process;

	-- generate the output signals
	outputGenerator: process(clock, outEnableLatch, delayDiv2Reg, useDataInputReg, latchedData, div1, div2, div2Delayed)
		variable div2Selected: std_logic;
	begin
		-- select synchronous div1/div2 or delayed div2
		if delayDiv2Reg = '1' then
			div2Selected := div2Delayed;
		else
			div2Selected := div2;
		end if;
		
		-- if output is not enabled, all outputs are set to 0
		if outEnableLatch = '1' then
			if useDataInputReg = '1' then
				-- overwrite output with latched data
				out1 <= latchedData(0);
				out2 <= latchedData(1);
				clockOut <= latchedData(0) xor latchedData(1);
				
				-- rawClock output is divided clock signal in overwrite mode
				rawClockOut <= div1;
			else
				-- use the divided signals
				out1 <= div1;
				out2 <= div2Selected;
				clockOut <= div1 xor div2Selected;
				rawClockOut <= div1 xor div2Selected;
			end if;
		else
			out1 <= '0';
			out2 <= '0';
			clockOut <= '0';
			rawClockOut <= '0';
		end if;
	end process;
	
end architecture rtl;

